# -*- encoding: utf-8 -*-

from __future__ import print_function

import argparse
import sys

import productmd.compose
import kobo.rpmlib

from .copy_compose import copy_compose  # noqa


def look_for_build(compose, nvr):
    """
    Look for a build NVR. If nvr contains an epoch, it will be used to match.
    If epoch is not present, build with any epoch will match.
    """
    parts = kobo.rpmlib.parse_nvr(nvr)
    name = parts['name']
    version = parts['version']
    release = parts['release']
    epoch = parts['epoch'] or None

    for variant in compose.rpms.rpms:
        for arch in compose.rpms.rpms[variant]:
            for srpm_nevra in compose.rpms.rpms[variant][arch]:
                srpm = kobo.rpmlib.parse_nvra(srpm_nevra)
                if srpm['name'] == name and \
                   srpm['version'] == version and \
                   srpm['release'] == release and \
                   (epoch is None or srpm["epoch"] == epoch):
                    return True

    return False


def has_build(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-q", "--quiet", help="do not print any output", action="store_true"
    )
    parser.add_argument("COMPOSE", help="compose to check")
    parser.add_argument("NVR", help="build to look for")

    opts = parser.parse_args(args)

    compose = productmd.compose.Compose(opts.COMPOSE)
    compose_id = compose.info.compose.id

    if look_for_build(compose, opts.NVR):
        if not opts.quiet:
            print("Compose %s has build %s" % (compose_id, opts.NVR))
    else:
        if not opts.quiet:
            print("Compose %s does not have build %s" % (compose_id, opts.NVR))
        sys.exit(1)


def get_listing(rpms, variants=[], arches=[], pkg_arches=[], skip_nevras={}):
    """Create product listing suitable for using as prepopulate file in Pungi."""
    listing = {}

    for variant in rpms.rpms:
        if variants and variant not in variants:
            continue
        for arch in rpms.rpms[variant]:
            if arches and arch not in arches:
                continue
            to_skip = skip_nevras.get(variant, {}).get(arch, set())
            for srpm_nevra in rpms.rpms[variant][arch]:
                pkgs = []
                srpm_data = rpms.rpms[variant][arch][srpm_nevra]
                srpm_nvr = kobo.rpmlib.parse_nvra(srpm_nevra)
                if set(srpm_data.keys()) & to_skip:
                    # If something from the build should be skipped, whole
                    # build should be skipped.
                    continue
                for nevra in srpm_data:
                    nvra = kobo.rpmlib.parse_nvra(nevra)
                    if pkg_arches and nvra["arch"] not in pkg_arches:
                        continue
                    if srpm_data[nevra]["category"] != "binary":
                        continue
                    pkgs.append("%(name)s.%(arch)s" % nvra)

                if pkgs:
                    listing.setdefault(variant, {}).setdefault(arch, {})[
                        srpm_nvr["name"]
                    ] = sorted(pkgs)

    return listing


def get_modular_packages(compose):
    """Return a mapping from variant and arch to a set of modular package NVRAs."""
    mapping = {}
    for variant in compose.modules.modules:
        for arch in compose.modules.modules[variant]:
            for nsvc in compose.modules.modules[variant][arch]:
                module = compose.modules.modules[variant][arch][nsvc]
                mapping.setdefault(variant, {}).setdefault(arch, set()).update(module["rpms"])
    return mapping
