# -*- coding: utf-8 -*-

import os
import glob

import productmd


def _get_label_name(label):
    if not label:
        return None
    return label.split("-")[0]


def _sort_key(compose):
    # Composes should be sorted according to this tuple. This mimicks sorting
    # in productmd, but without a check for release name. This is safe to do
    # here, because we always sort composes that were first filtered to the
    # same short name.
    return (
        tuple(productmd.common.split_version(compose.info.release.version)),
        compose.info.compose.date,
        productmd.composeinfo.COMPOSE_TYPES.index(compose.info.compose.type),
        compose.info.compose.respin,
    )


def find_composes(patterns):
    """Given a list of path globs, find all composes in those paths.
    """
    result = []
    for pattern in patterns:
        for topdir in glob.glob(pattern):
            for i in os.listdir(topdir):
                compose_path = os.path.join(topdir, i)
                if os.path.islink(compose_path):
                    continue
                if not os.path.isdir(compose_path):
                    continue

                compose = productmd.compose.Compose(compose_path)
                try:
                    # Access the metadata to raise an error.
                    compose.info
                except Exception as ex:
                    continue
                result.append(compose)
    return result


def filter_composes(
    composes,
    release_short=None,
    release_major_version=None,
    release_version=None,
    base_product_short=None,
    base_product_version=None,
):
    """Filter composes according to arguments."""
    result = []

    for compose in composes:
        ci = compose.info
        if release_short and ci.release.short != release_short:
            continue
        if release_major_version and ci.release.major_version != release_major_version:
            continue
        if release_version and ci.release.version != release_version:
            continue
        if base_product_short and ci.base_product.short != base_product_short:
            continue
        if base_product_version and ci.base_product.version != base_product_version:
            continue
        result.append(compose)

    base_products = set()
    major_versions = set()
    for compose in result:
        ci = compose.info
        if ci.release.is_layered:
            base_products.add("%s-%s" % (ci.base_product.short, ci.base_product.version))
        else:
            base_products.add(None)
        major_versions.add(ci.release.major_version)

    if len(base_products) > 1:
        raise ValueError(
            "More than 1 base product found: %s" % ", ".join(sorted(base_products))
        )
    if len(major_versions) > 1:
        raise ValueError(
            "More than 1 major version found: %s" % ", ".join(sorted(major_versions))
        )

    return result


def filter_compatible_composes(composes, compose):
    """Return composes compatible with provided compose (same release etc.)"""
    return filter_composes(
        composes,
        release_short=compose.info.release.short,
        release_major_version=compose.info.release.major_version,
        base_product_short=compose.info.base_product.short,
        base_product_version=compose.info.base_product.version,
    )


def get_previous_version(version):
    """Determine previous version from the given one. The major version is
    never decremented though.

    >>> get_previous_version("1.2")
    "1.1"
    >>> get_previous_version("1.2.1")
    "1.2.0"
    >>> get_previous_version("1.2.0")
    1.1.0
    >>> get_previous_version("1.0")
    ValueError
    """
    version_split = version.split(".")
    pos = len(version_split) - 1
    while pos > 1 and version_split[pos] == "0":
        pos -= 1
    if pos <= 1 and version_split[pos] == "0":
        # The only non-zero position was actually the major version. We don't
        # want to decrement that.
        raise ValueError("No previous version for this major version.")
    version_split[pos] = str(int(version_split[pos]) - 1)
    return ".".join(version_split)


def filter_previous_release(composes, compose):
    """Return composes for the previous product release"""
    try:
        previous_version = get_previous_version(compose.info.release.version)
    except ValueError:
        return []
    return filter_composes(
        composes,
        release_short=compose.info.release.short,
        release_version=previous_version,
        base_product_short=compose.info.base_product.short,
        base_product_version=compose.info.base_product.version,
    )


def filter_same_release(composes, compose):
    """Return composes for the same product release"""
    return filter_composes(
        composes,
        release_short=compose.info.release.short,
        release_version=compose.info.release.version,
        base_product_short=compose.info.base_product.short,
        base_product_version=compose.info.base_product.version,
    )


def filter_compose_type(composes, compose_type):
    """Return composes matching given compose_type"""
    return [i for i in composes if i.info.compose.type == compose_type]


def filter_older_than(composes, compose):
    """Return composes older than given compose"""
    return [i for i in composes if _sort_key(i) < _sort_key(compose)]


def _get_latest(composes):
    return max(composes, key=_sort_key) if composes else None


def _get_latest_by_type(composes, type_, compose=None):
    if compose:
        composes = filter_older_than(composes, compose)
    composes = filter_compose_type(composes, type_)
    return _get_latest(composes)


def get_latest_production(composes, compose=None):
    """Return latest production compose"""
    return _get_latest_by_type(composes, "production", compose=compose)


def get_latest_nightly(composes, compose=None):
    """Return latest nightly compose"""
    return _get_latest_by_type(composes, "nightly", compose=compose)


def get_previous_ga(composes, compose=None):
    """Return latest RC for *previous* product release"""
    if compose:
        composes = filter_previous_release(composes, compose)
    composes = filter_compose_type(composes, "production")
    composes = [i for i in composes if _get_label_name(i.info.compose.label) == "RC"]
    return _get_latest(composes)


def get_latest_beta(composes, compose=None):
    """Return latest Beta for current product release"""
    if compose:
        if _get_label_name(compose.info.compose.label) == "Beta":
            return None
        composes = filter_same_release(composes, compose)
    composes = filter_compose_type(composes, "production")
    composes = [i for i in composes if _get_label_name(i.info.compose.label) == "Beta"]
    return _get_latest(composes)


def get_previous_milestone(composes, compose):
    """Return latest compose with a different milestone name"""
    if not compose.info.compose.label:
        return None
    composes = filter_same_release(composes, compose)
    composes = filter_older_than(composes, compose)
    composes = filter_compose_type(composes, "production")
    label_name = _get_label_name(compose.info.compose.label)
    composes = [
        i
        for i in composes
        if _get_label_name(i.info.compose.label) != label_name
    ]
    return _get_latest(composes)


def get_same_milestone_previous_version(composes, compose):
    """Return latest compose with the same milestone name, but lower major
    label version.
    """
    if not compose.info.compose.label:
        return None
    composes = filter_same_release(composes, compose)
    composes = filter_older_than(composes, compose)
    composes = filter_compose_type(composes, "production")
    label_name = _get_label_name(compose.info.compose.label)
    composes = [
        i
        for i in composes
        if i.info.compose.label_major_version != compose.info.compose.label_major_version
        and _get_label_name(i.info.compose.label) == label_name
    ]
    return _get_latest(composes)


def get_same_milestone_previous_respin(composes, compose):
    """Return latest compose with the same milestone name, same major label
    version, but lower minor label version (older respin).
    """
    if not compose.info.compose.label:
        return None
    composes = filter_same_release(composes, compose)
    composes = filter_older_than(composes, compose)
    composes = filter_compose_type(composes, "production")
    label_name = _get_label_name(compose.info.compose.label)
    composes = [
        i
        for i in composes
        if i.info.compose.label_major_version == compose.info.compose.label_major_version
        and _get_label_name(i.info.compose.label) == label_name
    ]
    return _get_latest(composes)


def get_related_composes(composes, compose):
    """Return composes related to the given one"""
    result = {}

    result["beta"] = get_latest_beta(composes, compose)
    result["ga"] = get_previous_ga(composes, compose)

    result["latest_production"] = get_latest_production(composes, compose)
    result["latest_nightly"] = get_latest_nightly(composes)

    if compose is None:
        compose = result["latest_production"]
    result["previous_milestone"] = get_previous_milestone(composes, compose)
    result["same_milestone_previous_version"] = get_same_milestone_previous_version(
        composes, compose
    )
    result["same_milestone_previous_respin"] = get_same_milestone_previous_respin(
        composes, compose
    )

    return result
