# -*- coding: utf-8 -*-

"""
The `minor_version` argument was deprecated in 0.1.27 and should be removed in 0.2.0.
"""

import os
import warnings
from itertools import groupby
import re
import productmd


class VersionType(object):
    MAJOR = 0
    MINOR = 1
    FULL = 2


class ComposeSymlink(object):
    def __init__(self, compose_path):
        self.compose = productmd.compose.Compose(compose_path)
        # The path points to the compose/ subdirectory.
        self.path = os.path.dirname(self.compose.compose_path)

    def _link(self, name):
        """Create a symbolic link with given name at the same location as the
        compose. Existing file at the given path will be deleted atomically.
        """
        temp = os.path.join(self.path, "..", ".%s.tmp" % name)
        symlink = os.path.join(self.path, "..", name)

        os.symlink(os.path.basename(self.path), temp)
        os.rename(temp, symlink)

    def get_latest_link_name(self, minor_version=None, version_type=VersionType.MAJOR):
        """Get name of latest- symlink for current compose."""
        if minor_version is not None:
            _deprecated_arg("minor_version", "version_type")
        if minor_version:
            version_type = VersionType.MINOR

        formats = {
            VersionType.MAJOR: "{0.release.major_version}",
            VersionType.MINOR: "{0.release.major_version}.{0.release.minor_version}",
            VersionType.FULL: "{0.release.version}",
        }
        version = formats[version_type].format(self.compose.info)

        fmt = "latest-{0.release.short}-{1}"
        if self.compose.info.release.is_layered:
            # Add info about base product
            fmt += '-{0.base_product.short}-{0.base_product.version}'
            if self.compose.info.base_product.type != "ga":
                # Add base product type only if it's not ga
                fmt += "-{0.base_product.type}"
        return fmt.format(self.compose.info, version)

    def print_latest_link(self, minor_version=None, version_type=VersionType.MAJOR):
        """Print a latest- symlink with configured version pattern."""
        if minor_version is not None:
            _deprecated_arg("minor_version", "version_type")
        if minor_version:
            version_type = VersionType.MINOR
        print(self.get_latest_link_name(version_type=version_type))

    def create_latest_link(self, minor_version=None, version_type=VersionType.MAJOR):
        """Create a latest- symlink with configured version pattern."""
        if minor_version is not None:
            _deprecated_arg("minor_version", "version_type")
        if minor_version:
            version_type = VersionType.MINOR
        self._link(self.get_latest_link_name(version_type=version_type))


def print_latest_link(compose_path, minor_version=None, version_type=VersionType.MAJOR):
    """Convenience function for easy printing of latest symlink."""
    if minor_version is not None:
        _deprecated_arg("minor_version", "version_type")
    if minor_version:
        version_type = VersionType.MINOR
    ComposeSymlink(compose_path).print_latest_link(version_type=version_type)


def create_latest_link(compose_path, minor_version=None, version_type=VersionType.MAJOR):
    """Convenience function for easy creation of latest symlink."""
    if minor_version is not None:
        _deprecated_arg("minor_version", "version_type")
    if minor_version:
        version_type = VersionType.MINOR
    ComposeSymlink(compose_path).create_latest_link(version_type=version_type)


def _deprecated_arg(arg, replace):
    warnings.warn(
        "%r argument is deprecated, use %r instead" % (arg, replace), DeprecationWarning
    )


def update_symlinks(compose_path, dry_run=False):
    """Update the symlinks of the minor and major version of the compose"""
    latest_symlink_re = re.compile(
        r'(?P<prefix>latest-.+?-)(?P<version>\d+(\.\d+)*)(?P<suffix>-.+)?'
    )

    def natsort(s):
        return [int(t) if t.isdigit() else t.lower() for t in re.split(r'(\d+)', s)]

    def ver_len(s):
        return s.count(".")
    # Find all latest symlinks in the directory.
    all_symlinks = filter(
        lambda x: os.path.islink(os.path.join(compose_path, x)),
        (x for x in os.listdir(compose_path) if latest_symlink_re.match(x)),
    )
    # Filter the ones with most version components.
    links = sorted(all_symlinks, key=ver_len, reverse=True)
    if not links:
        raise RuntimeError(
            "No latest symlink found in the directory %r" % compose_path,
        )

    to_create = {}

    for latest_link in list(next(groupby(links, key=ver_len))[1]):
        m = latest_symlink_re.match(latest_link)
        link_pattern = "%s%%s%s" % (m.group("prefix"), m.group("suffix") or "")
        version = m.group("version")
        for i in range(version.count(".")):
            shorter_version = version[:version.rfind(".")]
            link = link_pattern % shorter_version
            dest = link_pattern % version
            if link in to_create:
                if natsort(version) > natsort(to_create[link][0]):
                    to_create[link] = (version, dest)
            else:
                to_create[link] = (version, dest)
            version = shorter_version

    for link, dest in sorted(to_create.items(), reverse=True):
        # Use the temp file in order to overwrite the existing symlinks
        temp = os.path.join(compose_path, ".link.tmp")
        # Create the symlink for the version
        print("%s -> %s" % (os.path.join(compose_path, link), dest[1]))
        if not dry_run:
            os.symlink(dest[1], temp)
            os.rename(temp, os.path.join(compose_path, link))
